UNIT PIT;

(**********************************************************************

  Control the 8253 or 8254 Chip (Programmable Interrupt Timer)

**********************************************************************)


INTERFACE

const
  HTicks_per_second = 1193181;

Function GetPit0Count : word;  { read value of channel 0 (clock)       }
Function GetPit1Count : word;  { read value of channel 1 (ram refresh) }
Function GetPit2Count : word;  { read value of channel 2 (speaker)     }

Procedure SetPit0Mode  (Mode : byte; value:word);
Function  GetPit0Mode : byte;   { only possible with the 8254 ! }


Function GetPitType : byte;   { reprograms timer 0 !! }
const
  PIT_EMUL  = 0;      { PIT is faulty or emulated by OS }
  PIT_8253  = 1;      { PIT is an 8253 }
  PIT_8254  = 2;      { PIT is an 8254 }


IMPLEMENTATION



Function GetPit0Count : word; assembler;
asm
  pushf
  cli

  mov  al, $00
  out  $43, al
  in   al, $40
  xchg ah,al
  in   al, $40
  xchg ah,al

  popf
end;



Function GetPit1Count : word; assembler;
asm
  pushf
  cli

  mov  al, $02  {## ??? warum nicht 0 ? }
  out  $43, al
  in   al, $41
  xchg ah,al
  in   al, $41
  xchg ah,al

  popf
end;



Function GetPit2Count : word; assembler;
asm
  pushf
  cli

  mov  al, $00
  out  $43, al
  in   al, $42
  xchg ah,al
  in   al, $42
  xchg ah,al

  popf
end;



Procedure SetPit0Mode (Mode : byte; value:word);
begin
  if mode > 6 then exit;
  mode := mode shl 1;
  mode := mode + $30;
  asm
    pushf
    cli

    mov   al, byte ptr mode
    out   043h, al
    jmp   @1
@1: mov   cx, word ptr value
    mov   al, cl
    out   040h, al
    jmp   @2
@2: mov   al, ch
    out   040h, al

    popf
  end;
end;



Function GetPit0Mode : byte;
{ only possible with the 8254 ! }
begin
  port [$43] := $C2; { readback command }
  GetPit0Mode := (port [$40] and $E) shr 1;
end;


(*************************************************************)

{ detect type of PIT chip - 8253, 8254 or emulated
  based on Sample program #7 from pctim002.zip
  by K.Heidenstrom (kheidens@actrix.gen.nz)
  pascal version by K.Hartnegg hartnegg@iname.com
}


Function GetPitType : byte;

   Function GetPITValue (channel:byte) : word;
   var i : word;
   begin
     port[$43] := channel shl 6;
     i := port [$40+channel];
     i := port [$40+channel] shl 8 + i;
     GetPITValue := i;
   end;

const
  TestValue = $55AA;
  Backwards = lo(TestValue) shl 8 + hi(TestValue);

  EXP_CSTAT = $30;    { Expected counter status }

var
  port61 : byte;
  rdb_status1, rdb_status2 : byte;
  rdb_count1, rdb_count2  : word;
  pit_type : byte;
  i,j    : word;

label
  got_type;

BEGIN
  pit_type := PIT_EMUL;

  asm
    pushf
    cli
  end;

  { turn off speaker & set gate2 input to low }
  port61 := port[$61];
  port[$61] := port61 and $FC;

  { program channel 2 to mode 0, two bytes, binary }
  port[$43] := $B0;
  port[$42] := lo(TESTVALUE);
  port[$42] := hi(TESTVALUE);

  { wait until the value of counter 0 changes }
  i := GetPitValue (0);
  repeat
    j := GetPitValue(0);
  until i <> j;
  i := j;
  repeat
    j := GetPitValue(0);
  until i <> j;

  { read value from counter 2, test if readout is stable }
  i := GetPitValue(2);
  j := GetPitValue(2);
  { if not then the PIT is bad or emulated }
  if (i <> j) or (i <> TESTVALUE) then goto got_type;


  { readback command will reverse lo/hi flag on a 8053 }
  port [$43] := $C8;
  rdb_status1 := port[$42];
  rdb_count1  := port[$42];
  rdb_count1  := port[$42] shl 8 + rdb_count1;

  i := GetPitValue(2);

  { read again to fix hi/lo flag }
  port [$43] := $C8;
  rdb_status2 := port[$42];
  rdb_count2  := port[$42];
  rdb_count2  := port[$42] shl 8 + rdb_count2;

  j := GetPitValue (2);

  if (rdb_status1 <> EXP_CSTAT) and (rdb_status2  <> EXP_CSTAT) and
     (i = BACKWARDS) and (j = TESTVALUE) then
        pit_type := PIT_8253;
  if (rdb_status1 =  EXP_CSTAT) and (rdb_status2  =  EXP_CSTAT) and
     (i = TESTVALUE) and (j = TESTVALUE) then
        pit_type := PIT_8254;

got_type:

  asm
    popf
  end;

  GetPitType := pit_type;
end;



END.
