Program SPLAY;
{$M 30000, 1024, 150000 }

{ Max heap size should be (64k-8)*2 + Safety }

{$D+,L+,Y+} {+++ for Debug, else ---}
{$Q+,R+,S+} {+++ for Debug, else ---}

{$B-}    {"Short-Circuit" boolean evaluation on}
{$E-}
{$I+}

{We should use a rather large stack because of the
 nested mergesong >> valsong >> loadsong call }

{Last Updated: 28/7/99}

(* Drive status/capabilities functions: *)
{function DoorOpen : Boolean;   {is door open?}
{function Locked : Boolean;     {is door locked?}
{function NoDisc : Boolean;     {is drive empty?}
{function DrvBusy : Boolean;    {is drive busy? (mostly playing)}
{function InPause : Boolean;    {is audio in pause?}
{function AudVolume : Byte;     {what's the audio channels' volume?}
                                {*refers to the cd-rom's headphones output}
{function SuppAudVid : Boolean; {does it support audio/video tracks?}
{function SuppAudChn : Boolean; {does it support audio channels? (volume)}

(* Disc/audio current information: *)
{procedure GetDiscInfo(var Info : DiscRec);          {not neccessary!, user's}
{procedure GetDiscLen(var DLen {: TMSF}{);           {get disc length/tracks}
{procedure GetTrkLoc(var TLoc {: TMSF}{; Trk : Byte);{get track's location}
{procedure GetTrkLen(var TLen : TMSF; Trk : Byte);   {get length of track}
{procedure CurDiscPos(var DPos {: TMSF}{);            {get disc position}
{procedure CurTrkPos(var TPos {: TMSF}{);             {get track position}
{procedure GetDiscRem(var DRem : TMSF);              {get disc remaining time}
{procedure GetTrkRem(var TRem : TMSF);               {get track remain. time}

(* CD-ROM drive commands: *)
{procedure EjectDoor;               {eject tray out, if not locked}
{procedure InsertDoor;              {insert tray inside}
{procedure LockDoor;                {lock tray (no eject till reboot/unlock)}
{procedure UnlockDoor;              {unlock the tray}
{procedure SetVolume(Vol : Byte);   {set all audio channel's volume}

(* Play/pause/resume procedures: *)
{procedure PlayFrom(From : TMSF);           {play from specified till end}
{procedure PlayRange(From, Till : TMSF);    {play specified range}
{procedure PlayAmount(From, Amount : TMSF); {play from spec., amount spec.}
{procedure PlayTrack(TrkNum : Byte);        {play spec. track, till disc end}
{procedure PlaySingle(TrkNum : Byte);       {play spec. track, till track end}
{procedure PlayPrevTrk;                     {play pervious track to current}
{procedure PlayNextTrk;                     {play next track to current}
{procedure PlayReverse(Skip : TMSF);        {play from current - specified}
{procedure PlayForward(Skip : TMSF);        {play from current + specified}
{procedure PauseAudio;                      {pause current audio if playing}
{procedure ResumeAudio;                     {resume audio, only if paused}

(* Utility functions, for both user, and other procedures: *)
{procedure SetTMSF(var Dest {: TMSF}{; T, M, S, F : Byte); {generate TMSF}
{procedure AddTMSF(var Dest {: TMSF}{; Src : TMSF);        {2 TMSF's sum}
{procedure SubTMSF(var Dest {: TMSF}{; Src : TMSF);        {2 TMSF's diff.}
{function CompTMSF(FstTime, SndTime : TMSF) : Byte;       {2 TMSF's relation}
{procedure DiscTime(var Dest : TMSF);                     {track to disc time}
{procedure TrkTime(var Dest : TMSF);                      {disc to track time}
{function SameDisc(var Dest, Src {: DiscRec}{) : Boolean;  {compare 2 DiscRec}

Uses CRT, CDAudio, SONGS, TimerInt, DOS;

Const
   Copyright = 'Song Play  Version 1.0  Copyright (c) 1997-1999 Idan Nof';

   NOE = 12; {Number of entries}
   DecLine: Array[1..NOE] of Byte = (3,5,6,8,9,11,12,14,15,17,18,20);
   Entry: Array[1..NOE] of Char=(#27,'P','E','D','S','Z','!','?','C','M','I','R');
   Menu: Array[1..NOE,1..2] of String[80]=
   (
   (' <ESC> = Exit', 'Exit to DOS.'),
   (' <P>   = Play Song', 'Play Current Song.'),
   (' <E>   = Edit Song', 'Edit Current Song.'),
   (' <D>   = Discs List', 'List Discs from "'+DiscsFName+'".'),
   (' <S>   = Songs List', 'List Songs from "'+SongsFName+'".'),
   (' <Z>   = Delete Disc', 'Delete Disc entry in "'+DiscsFName+'".'),
   (' <!>   = Delete Song', 'Delete Song entry in "'+SongsFName+'".'),
   (' <?>   = Search Song', 'Search Song entry in "'+SongsFName+'".'),
   (' <C>   = Change Active Song', 'Choose other song from current disc.'),
   (' <M>   = Import External Discs Data', 'Import an external discs file to "'+DiscsFName+'".'),
   (' <I>   = Import External Songs Data', 'Import an external songs file to "'+SongsFName+'".'),
   (' <R>   = Rescan Disc in drive, or add to database' ,'Rescan CD drive and add if not in database.')
   );

   NoSong: Boolean = True;
   DefaultPause = 100;
   Def_TFile_BufSize = 2000; { 2kb }

   Songs_BufSize : LongInt = 0;
   Songs_BufPtr : Pointer = Nil;

Var
  {SB Var}
  MixerAddr,MixerData: Word;
  Mix4_CD_LVol, Mix4_CD_RVol: Word;
  {SB Var}

  DiscsFile: IdxRecFile{File of IDXRec};

  Disc   : IDXRec;
  Song   : SongRec;

  OneMS  : Word;
  Key    : Char;
  Line:  String;
  Left: LongInt;
  Mark: 0..NOE;
  Valid: Array[1..NOE] of Boolean;
  Cache: SongID;
  Long, CD, SB, Timer_Handler: Boolean;
  Match, CurTrack: Byte;
  IntVec  : Pointer;

  SongsFile: Text;

  OldExit: Pointer;

Procedure VerifyBuffer_Exit; Far;
{ Flushes Songs File buffer (if in FmOutput mode) }
{ If previously allocated memory for SETTEXTBUF, dispose it }
Begin
   { If buffer is actually assigned & allocated }
   If Songs_BufSize > 0 then
      Begin
         { If file is open in output mode the Flush it }
         { (For example unexpected end of run while writing to the file) }
         If TextRec(SongsFile).Mode=FmOutput then Flush(SongsFile);
         { Free Memory }
         FreeMem(Songs_BufPtr, Songs_BufSize);
         Songs_BufSize := 0;
      End;
End;

Procedure Dispose_Buffer; Far;
Begin
   ExitProc := OldExit;
   VerifyBuffer_Exit;
End;

Procedure Finish(Good: Boolean);
{Notice that calling Finish(Good) is only at end of program }
Begin
   VerifyBuffer_Exit;

   TextColor(FG_Color);
   CursorOn;
   ClrScr;

   If CD and Locked then
      UnlockDoor;

   If Not Good then
      Begin
         Writeln('Quiting.');
         Halt;
      End;
End;

Function Check(What: Boolean; Critical: Boolean): Boolean;
Begin
    If What then Writeln('OK.')
       Else
          Begin
             Writeln('Failed !');
             If Critical then Finish(False);
          End
End;

Procedure FlushCache;
Begin
   With Cache do
        Begin
           Idx     :=0;
           BandN   :='No Band Name In Cache';
           SongN   :='No Song Name In Cache';
           Location:=-1;
        End;
End;

Procedure VDiscs;
Const _EOF = #26;
Var Temp: IdxRec;
    i: Byte;
Begin
   If (Exist(DiscsFName)) and (SizeFile(DiscsFName)=0) then
      Erase(DiscsFile);

   If Not Exist(DiscsFName) then
          Begin
             Write('Creating a new Discs File...');
             ReWrite(DiscsFile);
             Check(Exist(DiscsFName),True);

             {Clear bullshit }
             FillChar(Temp,SizeOf(Temp),0);
             With Temp do
                  Begin
                     Name:=Discs_Sig; {Signature}
                     {Length(Band)=ShortStr_Size should be > 28 (max size)}
                     Band:='Created ' + DateStr + ' ' + TimeStr + _EOF;

                     {Save some info about file structure }
                     With DRec.Length do
                          Begin
                            Frm:=Hi(SizeOf(IdxRec));
                            Sec:=Lo(SizeOf(IdxRec));
                            Min:=TracksPerDisc;
                            Trk:=ShortStr_Size;
                          End;
                  End;

             Write(DiscsFile,Temp);
             Close(DiscsFile);
          End
        Else
           Begin
              Write('Validating Discs File Header... ');
              Check(ValIdx(DiscsFile),True);
           End;
End;

Procedure VSongs(Verbose: Boolean);
Begin
   If Not Exist(SongsFName) then
      Begin
         Write('Creating an empty songs file...');
         ReWrite(SongsFile);
      End
   else
     If Verbose then Write('Validating Songs File...');

    { Validate & Close Songs File }
   If ValSong(SongsFile,TSongs) then
      Begin
         If Verbose then Writeln('OK.')
      End
        Else
           Begin
              Writeln;
              Writeln('The file "'+SongsFName+'" is corrupted and cannot be interpreted properly.');
              Writeln('Please repair the file (MANUALLY) and restart SPLAY.');
              Halt(0);
           End;
End;


Procedure InitFiles;
Begin
   Assign(DiscsFile,DiscsFName);
   Assign(SongsFile, SongsFName);

   { Calculate memory allocation for the Songs File's buffer }
   Songs_BufSize := MaxAvail;

   If Songs_BufSize >= MaxBufSize then
      Songs_BufSize:=MaxBufSize
   else If Songs_BufSize <= Def_TFile_BufSize then
           Songs_BufSize := 0;

   If Songs_BufSize>0 then
      Begin
        { Allocate memory }
        GetMem(Songs_BufPtr,Songs_BufSize);
        { Assign the buffer }
        SetTextBuf(SongsFile, Songs_BufPtr^, Songs_BufSize);

        { Modify our exit proc. to flush & dispose the file's buffer }
        OldExit := ExitProc;
        ExitProc := @Dispose_Buffer;

      End;

   VDiscs;
   VSongs(True);
   Pause;
End;

Procedure InitSB;
Begin
   If Not SB then Exit;

   DetectSBPort;
   MixerAddr :=BaseIO+$04;    {Mixer Address (register select) Port}
   MixerData :=BaseIO+$05;    {Mixer Data Port}

   Mix4_CD_LVol := $36;
   Mix4_CD_RVol := $37;

   { Reset DSP - Check for existence of a card at 220h }
   SB := (ResetDSP=$AA) and (GetDSPVer>=4.00);
End;

Procedure ShowTrackTime; Far;
Function LeadZero(b : Byte): String;
begin
   If B<10 then
      LeadZero := '0'
   else
      LeadZero := '';
end;

Var Time: TMSF;
    Save, WX, WY: Byte;
Begin
   Save:=TextAttr;
   WX:=WhereX;
   WY:=WhereY;

   TextColor(FG_Color);
   GotoXY(55,24);
   GetTrkRem(Time);

   With Time do
        Write('<Track ',LeadZero(Trk),Trk,', Left ',LeadZero(Min),Min
        ,':',LeadZero(Sec),Sec,'>');

   TextColor(FGDis_Color);
   Write('---');

   TextAttr:=Save;
   GotoXY(WX,WY);
End;

Procedure NewInt9; Interrupt;
Const UpArrow=72; DownArrow=80;

Var Key: Byte;
    Less: Boolean;

Procedure ChangeVol(Source: Byte);
{Takes a 5-bit Source and Increases/Decreases it in 1 }
Const
    MaxVol = 31; {2^5-1}

Var OldVol, Def: Integer;
Begin
   Port[MixerAddr]:=Source;
   {Take 5 MSB bits }
   OldVol := Port[MixerData] and $F8 {1111 1000};
   OldVol := OldVol Shr 3 {1111 1000 >> 0001 1111};

   If Less then Def:=-1
           else Def:=1;

   OldVol:=OldVol+Def;

   If OldVol in [0..MaxVol] then
      Begin
         OldVol := OldVol Shl 3;
         {Same Source as before ...}
         Port[MixerData]:=Byte(OldVol);
      End;

   {
    SBPRO - 0x28h         (4 bits-per-channel)
    SB16  - 0x36h , 0x37h (5 bits-per-channel)

    Only SB16 + (DSP 4.xx+) is supported
   }

End;

Begin
   Asm
     PUSHF  { Save Flags }
     CLI    { Disable Interrupts }
   End;
   Key := Port[$60];
   If Key in [UpArrow,DownArrow] then
      Begin
         Less := (Key=DownArrow);
         ChangeVol(Mix4_CD_LVol);
         ChangeVol(Mix4_CD_RVol);

         Port[$20]:=$20; { Signal End Of Interrupt to the Intel 8??? Chip}
      End
   else
      Asm
       PUSHF             { Save flags                }
       CALL OldInt9      { Call the old INT 9        }
      End;

   Asm
     POPF { Restore Flags }
   End;
End;

Function ChangeTrack(Trk: Byte): Boolean;
Var T: Word;
Begin
   If SearchSongName(SongsFile,Cache,Song.Band,Disc.Songs[Trk]) then
      Begin
         LoadSong(SongsFile,Song,Cache,True,T);
         CurTrack:=Trk;
         {NoSong=False ...}
         ChangeTrack:=True;
      End
   else
      ChangeTrack:=False;
End;

Procedure Play(Var Song: SongRec; Edit: Boolean);
Var Wor:WordRec;
    Cnt,Line: Integer;
    Heb, ShouldPlay, NoInfo, Move, MoveLine, OweYou: Boolean;
    CurLoc: Loc;
    Contin, T: Integer;
    SavedX, SavedY, FirstLetter: Byte;
    Tem: Word;
    PauseCount: Byte;
    SongWords: String;
    KeyTransfer: Boolean;

Procedure UpdateDown;
Begin
   GotoXY(1,25);
   TextColor(Magenta);
   Write('E<X>it, <P>ause');
   If SB then Write(', Volume Up/Down: <Up>/<Down>');

   If (CD) and (UP(Song.Band)=UP(Disc.Band)) then
      Write(', Prev/Next Track: <->/<+>');

   If NoInfo then
      Begin
         TextColor(FG_Color);
         Write(' No Info');
      End;

   ClrEol;

End;

Procedure ShowWords;
Var Count: Byte;
Begin
    TextColor(SongBody_Color);
    With Song do
         For Count:=Contin to T do
             Begin
                If Heb then Write(' ':(80 - (Length(Words[Count])) ) );
                Write(Words[Count]);
                If (Not Heb) and (Count<T) then Writeln;
             End;

    With Song do
     If T < Lines then
        Begin
           Long:=True;
           Contin:=T+1;
           If Lines-T < 22 then
                              T:=Lines
                           else
                              Inc(T,21);

           If Words[Count]='' then
                                Begin
                                   GotoXY(WhereX,WhereY-1);
                                   Dec(Count);
                                End;
           If Heb then
                     Begin
                         GotoXY((80 - (Length(Words[Count])+10) ), WhereY-1);
                         Write('(...) ');
                     End
                  else
                     Write(' (More...)');

       End
    else
       If T = Lines then
          Begin
              Long:=False;
              If Heb then
                        Begin
                            GotoXY((80 - (Length(Words[Count])+6) ), WhereY-1);
                            Write('() ');
                        End
                     else
                        Write(' (End)');
          End;
    End;

Procedure FirstTime;
Var i: Byte;
    Rem1,Rem2: TMSF;

Begin
   ClrScr;
   TextColor(SongHead_Color);
   If Heb then Write(' ':( 80-(Length(Song.Name)+Length(Song.Band)+3) ) ) ;
   Write(Song.Name,' / ',Song.Band);
   If Not Heb then Writeln;

   If Heb then Write(' ':( 80-(Length(Song.Name)+Length(Song.Band)+3) ) );
   For i:=1 to Length(Song.Name)+Length(Song.Band)+3 do Write('-');
   If Not Heb then Writeln;

   Long:=False;
   Contin:=1;
   T:=21;
   If Song.Lines < 21 then T:=Song.Lines;

   ShowWords;

   {Update down }
   GotoXY(1,24);
   TextColor(FGDis_Color);
   Writeln('-------------------------------------------------------------------------------');
   GotoXY(2,24);
   TextColor(Cyan);
   Write('Disc: ');
   Write(Disc.Name);
   If Up(Strip(Disc.Name)) <> Up(Strip(Song.Album)) then
      Begin
         GotoXY(WhereX+1,WhereY);
         If WhereX+Length(Song.Album)+2 < 80 then Write('(',Song.Album,')');
      End;
   Writeln;

   TextColor(Magenta);

   {Song.Start}
   If Edit then
      Write('<P>ause')
   else {Ready}
      Write('E<X>it, <P>ause');

   {Update down}

   With Song do
    If CD and (Edit or Ready) then
       Begin
          {This baby should bypass the !@#% waking time
           of a sleepy CDROM.
           On the other hand, it might not work at all ...}

          GotoXY(50,25);
          FlushKeyBuffer;
          Write('Please wait...');

          PlaySingle(CurTrack);

          CurDiscPos(Rem1);
          Repeat CurDiscPos(Rem2);
                 {A delay should be here ? }
          Until (CompTMSF(Rem1,Rem2)<>SameTime) or (KeyPressed);
          GotoXY(50,25);
          ClrEol;

          ShouldPlay:=True;
          If Timer_Handler then Install_Handler(@ShowTrackTime, 2);
          {2Hz just to be sure we're doing it fast enough}
       End
    else
       ShouldPlay := False;

   If Not (Edit or Song.Ready) then
      Begin
        Song.Start:=0;
        NoInfo:=True;
      End
   else
      NoInfo:=False;

   FlushKeyBuffer;
   GotoXY(59,25);
   With Song do
        If Edit then
           Begin
              { Clear Crap }
              FillChar(Pause,SizeOf(Pause),0);
              Write('Start: ');
              Start:=DelaySS(67,25, Key, ShouldPlay, KeyTransfer);
           End
        else
           Begin
              Write('To Start: ');
              Left:=DelayS(Start,69,25,Key, ShouldPlay, KeyTransfer);
           End;

   GotoXY(59,25);
   ClrEol;

   FlushKeyBuffer;
   UpdateDown;
End;

Procedure GetWord;
Function RtL_SpacePos(S: String): Byte;
{ Find Space, from Right to Left }
Var i: Byte;
Begin
    For i:=Length(S) downto 1 do
        If S[i]=' ' then Break;

    If S[i]=' ' then
       RtL_SpacePos:=i
    else
       RtL_SpacePos:=0;
End;

Function LtR_SpacePos(S: String): Byte;
{ Find Space, from Left to Right }
Var i: Byte;
Begin
    For i:=1 to Length(S) do
        If S[i]=' ' then Break;

    If S[i]=' ' then
       LtR_SpacePos:=i
    else
       LtR_SpacePos:=0;
End;

Var Temp: String;
Begin
    Temp:= SongWords;

    With Song do
         If Edit then
            Begin
               If (PauseCount<>1) and (Wor.Words<>'') then
                  Pause[Line,PauseCount-1]:=Left;
            End
         else With Wor do
              If Ready then
                 P:=Pause[Line,PauseCount]
              else
                 P:=DefaultPause; {Default Pause Length}

    If Heb then Wor.Len:=RtL_SpacePos(Temp)
           else Wor.Len:=LtR_SpacePos(Temp);

    If (Length(Wor.Words)>0) or (MoveLine) then
       Begin
           MoveLine:=False;

           If Heb then
              Begin
                 If Wor.Len=0 then SongWords:=''
                 else Delete(SongWords,Wor.Len,Length(SongWords));
                 Delete(Temp,1,Wor.Len);
              End
           else
              Begin
                  Delete(SongWords,1,Wor.Len);
                  If Wor.Len>0 then Delete(Temp,Wor.Len,Length(Temp))
                               else SongWords:='';
              End;

           Wor.Words:=Temp
       End
         else
            Begin
               MoveLine:=True;
               Inc(Line);
               Wor.Words:=SongWords;
               SongWords:=Song.Words[Line];
            End;
End;

Label 1;

Var Ch: Char;
    Count: Integer;
    ExitEvent : Boolean;
    Rem1: TMSF;

Begin
   If SB then InstallKBDHandler(@NewInt9);{*Install Keyboard Handler*}

1:
   Key:=#0;
   KeyTransfer:=False;
   PauseCount:=1;
   Move:=False;
   Cnt:=0;
   ExitEvent := False;
   Heb:=Song.Hebrew;     {Heb is the same}
   OweYou:=False;

   If Heb then
      FirstLetter:=82
   else
      FirstLetter:=1;

    Line:=1;
    SongWords := Song.Words[1];
    Left:=0;

    CursorOff;
    FirstTime;

    GetWord;
    With CurLoc do
         Begin
             Y:=3;
             If Heb then X:=81-(Length(Wor.Words))
                    else X:=1;
         End;

    Repeat
        If KeyTransfer or KeyPressed then
           Begin
              If KeyTransfer then KeyTransfer:=False
                             else Key:=UpCase(ReadKey);

              Case UpCase(Key) of
                   #0: Ch:=ReadKey; {Read Func Key}

                   {* Note: FlushKeyBuffer doesn't handle
                            function keys (#0 + #??) who's #0 was already
                            read (cause the #?? is also out of the buf)
                   }

                   'X':
                      Begin
                         {Audio Paused before}
                         GotoXY(1,25);
                         ClrEOL;
                         GotoXY(30,25);
                         TextColor(11);
                         Write('Exit (Y/N) ?');
                         CursorOn;
                         If Cnt=0 then Key:=Yes else
                            Repeat
                                Key:=UpCase(ReadKey);
                                If Key=#0 then ReadKey;
                            Until Key in YesNo;

                         CursorOff;
                         FlushKeyBuffer;
                         UpdateDown;
                         GotoXY(CurLoc.X,CurLoc.Y);
                         If Key=Yes then
                            ExitEvent := True
                         else
                            If ShouldPlay and InPause then ResumeAudio; {Audio paused earlier}
                      End;

                     'P':
                        Begin
                           If (ShouldPlay) and (Not InPause) then PauseAudio;
                           GotoXY(1,25);
                           ClrEOL;
                           GotoXY(25,25);
                           TextColor(11);
                           Write('Press Any Key to Continue');
                           CursorOn;
                           Pause;
                           CursorOff;
                           UpdateDown;
                           GotoXY(CurLoc.X,CurLoc.Y);
                           If ShouldPlay and InPause then ResumeAudio;
                        End;

                 '-','+': If (CD) and (UP(Song.Band)=UP(Disc.Band)) then
                  {If disc in drive AND playing a song from the current disc }
                        Begin
                           If Key='+' then Count := CurTrack+1
                                     else Count := CurTrack-1;

                           With Disc.Drec.Length do
                                If Count=0 then Count:=Trk
                                   else If Count=Trk+1 then Count:=1;

                           If ChangeTrack(Count) then
                              Begin
                                If ShouldPlay then Remove_Handler(@ShowTrackTime);
                                Goto 1;
                              End;
                        End;
                End; {Case }
            End
         Else
            Begin
               If Not OweYou then
                  Begin
                     Inc(Cnt);
                     Inc(PauseCount);
                  End;

               If Long and (Line=Contin) then
                  Begin
                     Move:=False;
                     Window(1,3,80,23);
                     ClrScr;
                     Window(1,3,80,24);
                     ShowWords;
                     Window(1,1,80,25);

                     GetWord;
                     With CurLoc do
                          Begin
                             Y:=3;
                             If Heb then X:=81-(Length(Wor.Words))
                                    else X:=1;
                          End;
                  End;

               If Not OweYou then
                  Begin
                     TextColor(WordPlay_Color);
                     With CurLoc do GotoXY(X,Y);
                     Write(Wor.Words);
                  End;

               If (Not MoveLine) and (Length(Wor.Words)>0) then
               If Edit then
                         Begin
                            If OweYou then Left:=Left+DelayMSS(Key,ShouldPlay, KeyTransfer)
                                      else Left:=DelayMSS(Key,ShouldPlay, KeyTransfer);
                         End
                       else
                         Begin
                            If OweYou then Left:=Left+DelayMS(Wor.P,Key,ShouldPlay, KeyTransfer)
                                      else Left:=DelayMS(Wor.P, Key, ShouldPlay, KeyTransfer);
                         End;

               OweYou := (KeyTransfer) and (Key in ['X','P']);
               If OweYou then Continue;

               TextColor(FG_Color);
               With CurLoc do GotoXY(X,Y);
               Write(Wor.Words);

               {Line=Line, Word=Cnt}
               {Blank Lines Count!!!!!!!!!!!!!@!@!@!@!}

               GetWord;

               With CurLoc do
                    Begin
                       If MoveLine then
                          Begin
                             Inc(Y);
                             X:=FirstLetter;
                             Move:=True;
                             PauseCount:=0;
                             Dec(Cnt,2);
                          End
                      else
                         If Heb then
                                   Dec(X,Length(Wor.Words)+1)
                                   else
                                     Begin
                                         X:=WhereX+1;
                                         If Move then
                                                   Begin
                                                     Dec(X);
                                                     Move:=False;
                                                   End;
                                     End;

                     End;
            End;
    Until (Line>=Song.Lines+1) or ExitEvent;

    If ShouldPlay and DrvBusy then
       Begin
          With Rem1 do
               Repeat GetTrkRem(Rem1);
               Until ((Min=0) and (Sec<1)) or (KeyPressed);
          If Keypressed then FlushKeyBuffer;
          While (DrvBusy) and (Not keypressed) do;
          If Keypressed then FlushKeyBuffer;
          If DrvBusy then PauseAudio;
       End;

    If ShouldPlay and Timer_Handler then
       Remove_Handler(@ShowTrackTime);

    TextBackground(Black); {Some color re-initializations}
    TextColor(FG_Color);

    If (Not ExitEvent) and (Edit) then
       Begin
         {Modify Songs File according to changes made}
         SaveSong(Song,Cache);

         {Check out if we didn't corrupt the file (stupid us) }
         VSongs(False);

         {Reload modified song from file (just to be sure ...) }
         Line := Cache.Idx;
         FlushCache;
         If SearchSongIdx(SongsFile,Cache,Line) then
            LoadSong(SongsFile,Song,Cache,True,Tem);
       End;

    If (Not ExitEvent) and (CD) and (UP(Song.Band)=UP(Disc.Band)) then
       {If disc in drive AND playing a song from the current disc }
       Begin
          Count := CurTrack+1;
          If Count<=Disc.Drec.Length.Trk then
             Begin
                ChangeTrack(Count);
                Goto 1;
             End;
       End;

    If SB then RestoreKBDHandler;
End;

Procedure InitCD;
Var
  SavedX, SavedY: Byte;
Begin
   Song.Name := 'No Song Loaded';
   FlushCache;

   If Not CD then Exit;
   TextBackground(BG_Color);
   TextColor(FG_Color);

   ClrScr;
   CursorOn;
   If CD AND (MSCDEXVer<2.10) {If not loaded then MSCDEXVER=0.00} then
      Begin
         CD:=False;
         Exit;
      End;

   If DoorOpen then
      Begin
         Write('Closing Tray... ');
         InsertDoor;
         SavedX:=WhereX;
         SavedY:=WhereY;
         While (DrvBusy or DoorOpen) and (Not Keypressed) do
               Begin
{                  GotoXY(15,15);
                  Write('DrvBusy: ',DrvBusy,'; DoorOpen: ',DoorOpen);}
                  GotoXY(SavedX,SavedY);
                  Write('CD-ROM busy (any key to quit) ...');
               End;
         Check(Not DoorOpen,False);
      End;

   If NoDisc then
      Begin
         Writeln('No Disc In Drive');
         Writeln('----------------');
         Writeln('Insert audio disc now or press any key to continue without CD-support');
         FlushKeyBuffer;
         Write('Status : Door Is ');
         Write('    ed'); {openED , closED}

         GotoXY(WhereX-6,WhereY);
         Repeat
             If DoorOpen then
                            Write('Open')
                         else
                            Write('Clos');
             GotoXY(WhereX-4,WhereY);
             Long:=DoorOpen;
             While Long=DoorOpen do If Keypressed then Break;
         Until (Keypressed) or (Not NoDisc);
         FlushKeyBuffer;
         If NoDisc then CD:=False
                   else CD:=True;
         Writeln;
      End
   else CD:=True;

   If CD and NoDisc then
                       CD:=False;

   If NOT CD then Exit; {Exit Current Procedure}

   Writeln;

   If Not Locked then
          Begin
             Write('Locking Tray... ');
             LockDoor;
             Check(Locked,False);
          End;

   If DrvBusy then
      Begin
         Write('Drive Busy, Trying to Pause ... ');
         PauseAudio;
         Check(Not DrvBusy,True);
      End;
End;


Procedure LoadCorrect;
{ Counts on Disc }
Var i: Integer;
    T:  Word;
Begin
   With Disc do
        Begin
          Writeln('Matches Disc: "',Name,'"');
          For i:=1 to Drec.Length.Trk do
              Writeln('Track #',i,' : ',Songs[i]);
          Write('Searching For Valid Songs...');
          Match := SearchSongList(SongsFile, Band, Songs, CurTrack, i
                   , Drec.Length.Trk);
          Writeln('Matched ',Match,' out of ',Drec.Length.Trk,' songs.');
          If Match > 0 then
             Begin
                NoSong:=False;
                If SearchSongName(SongsFile, Cache, Band, Songs[1]) then
                   Begin
                      LoadSong(SongsFile,Song,Cache,True,T);
                      CurTrack:=1;
                   End
                else
                   Begin
                      SearchSongIdx(SongsFile, Cache, i);
                      LoadSong(SongsFile,Song,Cache,True,T);
                   End;
             End;
         End;
End;

Procedure CheckCD;
Var Same : Boolean;
    Heb: Byte;

Procedure InsertCD;
Var Count: Integer;
Begin
    Writeln('not found.');
    Write('Do you want to insert this disc to the database(Y,N) ?');
    CursorOn;
    FlushKeyBuffer;
    Repeat
         Key:=UpCase(ReadKey);
         If Key=#0 then ReadKey;
    Until Key in YesNo;

    If Key=Yes then
       Begin
          Writeln(' Y');
          Same:=True;
          Heb:=AskLang;
          If Heb=0 then Exit;

          GotoXY(1,1);
          With Disc.DRec do
               Writeln('Total Time ',Length.Min,':',Length.Sec,', Total Tracks ',Length.Trk);

          CursorOn;
          If Heb=1 then
                     Begin
                       Disc.Name:=Hebrew(' : ');
                       GotoXY(1,4);
                       Disc.Band:=Hebrew(' : ');
                    End
                 else   {Key='E'}
                    Begin
                       Write('Disc Name: ');
                       Readln(Disc.Name);
                       GotoXY(1,4);
                       Write('Band Name: ');
                       Readln(Disc.Band);
                     End;

           ClrScr;
           GotoXY(1,2);

           With Disc do
           For Count:=1 to Drec.Length.Trk do
               Begin
                   Str(Count, Line);
                   If Heb=1 then Songs[Count]:=Hebrew(' : #'+Line+' ')
                          else
                             Begin
                                 Write('Song #',Count:2, ' : ');
                                 Readln(Songs[Count]);
                             End;
               End;
           Reset(DiscsFile);
           Seek(DiscsFile, FileSize(DiscsFile)); {Goto EOF}
           Write(DiscsFile, Disc);
           Writeln('Entry ', FileSize(DiscsFile)-1,' successfully inserted.');
           Close(DiscsFile);

           {Check out to see if we didn't corrupt the file ... }
           VDiscs;
       End
    Else
       Begin
           Writeln(' N');
           Disc.Name:='Unknown Disc';
           Song.Name:='No Song Selected';
           NoSong:=True;
       End;
    CursorOff;
End; {InsertCD}

Var Temp: IdxRec;
    Count: Integer;

Begin
    Count:=0;
    Match:=0;
    Same:=False;

    GetDiscInfo(Disc.Drec);
    Writeln;
    With Disc.Drec do
             If Track[1].Trk=1 then {Data Track}
                Begin
                    {Assume Data disc if first track is data }
                    Writeln('Data Discs aren''t supported !');
                    CD:=False;
                End
             else With Length do
                Begin
                   Writeln('Total Time ',Min,':',Sec,', Total Tracks ',Trk);
                   If Trk > TracksPerDisc then
                      Begin
                         Writeln('Error: Max ',TracksPerDisc, ' Tracks Per Disc');
                         Writeln('       All tracks past the maximum will be truncated.');
                         Writeln('       Please contact the author for a fix.');
                         Writeln;
                         Trk:=TracksPerDisc;
                      End;
                End;


    If CD then
       Begin
          Write('Searching Discs Database... ');
          {Current Disc in drive = Disc.Drec}
          Reset(DiscsFile);

          If FileSize(DiscsFile) > 1 then
             Begin
                Seek(DiscsFile,1);  {Skip header record }
                Repeat
                    Inc(Count);
                    Read(DiscsFile, Temp);
                    Same:=SameDisc(Disc.Drec,Temp.Drec)
                Until (Eof(DiscsFile)) or (Same) or (Count=FileSize(DiscsFile));
             End
          else Same:=False;

          Close(DiscsFile);

          {InsertCD updates variables Same & Disc}
          If Not Same then
             InsertCD
          else
            Disc := Temp; {!!}

          If Same then LoadCorrect;
       End;

End;

Procedure Show(Item: Byte);
Begin
   If Valid[Item] then TextColor(Menu_Color)
                  else TextColor(FGDis_Color);
   Writeln(Menu[Item,1]);
End;

Procedure MarkSpot(Enabled: Boolean; Mark: Byte);
Begin
   TextBackground(LightBlue);
   If Enabled then TextColor(Menu_Color)
              else TextColor(FGDis_Color);
   GotoXY(1,DecLine[Mark]);
   Show(Mark);
   TextBackground(Black);
End;

Procedure UnMarkSpot(Enabled: Boolean; Mark: Byte);
Begin
    If Enabled then TextColor(Menu_Color)
               Else TextColor(FGDis_Color);
    GotoXY(1,DecLine[Mark]);
    Show(Mark);
End;

Procedure CheckDisc;
Var T: Word;
    Found : Boolean;
Begin
   If UP(Song.Name)=UP(Cache.SongN) {if changesong
      loaded sucessfully} then
   Begin
      T:=0;
      With Disc do
           Repeat
              Inc(T);
              Found:=(UP(Disc.Songs[T])=UP(Song.Name));
           Until (T>=Drec.Length.Trk) or (Found);

      If Found then
         Begin
            CurTrack:=T;
            NoSong:=False;
         End;
   End;
End;
(* ------------------------------------------------------------------------ *)
Procedure ProcessParameters;
Procedure ShowHelp;
Begin
   Writeln('Available Parameters: ');
   Writeln;
   Writeln('/D - Disable CD Support');
   Writeln('/S - Disable SB Support (No INT 9 Handler)');
   Writeln('/T - Disable CD Time-Left (No INT 8 Handler)');
   Writeln;
   Halt(1);
End; {ShowHelp}

Var i: Integer;
    S: String;

Begin
   If ParamCount>0 then
      For i:=1 to ParamCount do
          Begin
             S := ParamStr(i);
             If (S[1]='/') and (Length(S)=2) then
                Case UpCase(S[2]) of
                     'D': CD:=False;
                     'S': SB := False;
                     'T': Timer_Handler := False;
                     else
                        ShowHelp;
                End { Case }
             else ShowHelp;
          End;

End; { ProcessParameters }

Procedure Init1;
{Var T: Word;}
Begin
   If Test8086<2 then
      Abort('Need an 80386-compatible system to run this program');

   NoSong:=True;
   CurTrack:=0;
   Writeln(Copyright);
   Writeln;

   { Default Values }
   CD:=True;
   SB:=True;
   Timer_Handler := True;

   ProcessParameters;

   InitCD;
   Writeln;
   Write('CD-Support ');
   If CD then Writeln('Enabled.')
         else Writeln('Disabled.');

   InitSB;
   Write('SB-Support ');
   If SB then Writeln('Enabled.')
         else Writeln('Disabled.');

   Writeln;
   InitFiles;
End; {Init1 }

Procedure Init2;
Var T: Word;
Begin
   If CD then
      Begin
         CheckCD;
         Pause;
      End;

   If NoSong then {Load first song in SongsFile}
      Begin
          If SearchSongIdx(SongsFile,Cache,1) then
             Begin
                LoadSong(SongsFile,Song,Cache,True,T);
                NoSong:=False;
             End
      End;

    {Esc}
    Valid[1]:=True;

    {Check Import External Discs File}
    Valid[10]:=True;

    {Check Import External Songs File}
    Valid[11]:=True;

    {Check Rescan/Add Disc in drive}
    Valid[12]:= (MSCDEXVer>=2.10) ;

    If Not CD then
           Begin
              Menu[NOE,1]:=' <R>   = Enable CD-Support & Rescan Disc in drive';
              Menu[NOE,2]:='Enable CD-Support (if possible) and Rescan CD drive.';
           End;

    Mark:=1;                {Initialize Cursor Pointer on [ESC]}
End; { Init2 }

{ Main program }
Const
    Run : Boolean = False;
    Ins : Boolean = False;

Var Found: Boolean;

Begin   (* main program *)
    Init1; { Init CD, SB, Files }
    Init2;

    Repeat
        {Check Play Song}
        Valid[2]:=Not NoSong;

        {Check Edit Song}
        Valid[3]:=Not NoSong;

        {Check Discs List}
        Valid[4]:=TDiscs>0;

        {Check Songs List}
        Valid[5]:=TSongs>0;

        {Check Delete Disc}
        Valid[6]:=TDiscs>0;

        {Check Delete Song }
        Valid[7]:=False{TSongs>0};

        {Check Search Song }
        Valid[8]:=TSongs>0;

        {Check Change Song }
        Valid[9]:=TSongs>1;

        ClrScr;
        TextColor(Choose_Color);
        Writeln('Choose :');
        Writeln('--------');
        Show(1);
        TextColor(Buf_Color);
        Writeln('*************');
        Show(2);
        Show(3);
        TextColor(Buf_Color);
        Writeln('******************');
        Show(4);
        Show(5);
        TextColor(Buf_Color);
        Writeln('*******************');
        Show(6);
        Show(7);
        TextColor(Buf_Color);
        Writeln('********************');
        Show(8);
        Show(9);
        TextColor(Buf_Color);
        Writeln('***************************');
        Show(10);
        Show(11);
        TextColor(Buf_Color);
        Writeln('***********************************');
        Show(12);
        TextColor(Buf_Color);
        Writeln('*************************************************');
        MarkSpot(True,Mark);    { Mark Current Spot }
        Repeat
            GotoXY(1,22);
            If Valid[Mark] then TextColor(Buttom_Color)
                           else TextColor(FGDis_Color);
            Writeln(Menu[Mark,2],' ':79-Length(Menu[Mark,2]));
            Writeln;
            TextColor(Info_Color);
            Writeln('=':(80-Length(Song.Name)) div 2,Song.Name,'=');
            If CD then
               Begin
                  Write('-':(80-Length(Disc.Name)) div 2,Disc.Name,'-');
                  TextColor(CDSupp_Color);
                  GotoXY(66,1);
                  Write('CD-SUPPORT MODE');
               End;
            TextColor(FG_Color);
            CursorOn;
            GotoXY(9,1);
            If Ins then
               Begin
                  Ins:=False;
                  Key:=Entry[Mark];
               End
            Else
               Key:=UpCase(ReadKey);

            CursorOff;
            Run:=True;      {!!!}
            Case Key of
                 #13:     {Enter}
                    Begin
                       Ins:=True;
                       Run:=False;
                    End;
                 #0:      {function key}
                   Begin
                      Key:=ReadKey;       {Function key out of kbd buffer}
                      FlushKeyBuffer;
                      Run:=False;
                      If Key in [#72{Up},#80{Down}] then
                         Begin
                            UnmarkSpot(Valid[Mark], Mark);
                            If Key=#72 then  {Up Arrow}
                                          If Mark=1 then
                                             Mark:=NOE
                                          else
                                             Dec(Mark)
                                        else  {Down Arrow}
                                           If Mark=NOE then
                                                         Mark:=1
                                                       else
                                                         Inc(Mark);
                            MarkSpot(Valid[Mark], Mark);
                         End;
                   End;

                 #27:                          {Esc}
                    Begin {Present only to keep run = false}
                    End;

                 'P':
                    If Valid[2] then
                                  Play(Song,False)  {Play Song}
                                else
                                  Run:=False;

                 'E':
                    If Valid[3] then
                                  Play(Song,True)   {Edit Song}
                                else
                                  Run:=False;

                 'D':
                    If Valid[4] then
                                  DiscsList(DiscsFile)  {Discs List}
                                else
                                   Run:=False;

                 'S':
                    If Valid[5] then
                                  SongsList(SongsFile,Cache,True){Songs List}
                                else
                                  Run:=False;

                 'Z':
                    If Valid[6] then
                                  Begin
                                     DelDisc(DiscsFile);{Delete Disc entry}
                                     CheckCD;           {Update current status}
                                  End
                                else
                                   Run:=False;

                 '!':
                    If Valid[7] then
                                   {DelSong(SongsFile)  {Delete Song entry}
                                else
                                   Run:=False;

                 '?':
                    If Valid[8] then
                       Begin
                          If SearchSong(SongsFile,Song,Cache) then
                             CheckDisc;
                       End
                    else
                       Run:=False;


                 'C':
                    If Valid[9] then
                                   Begin
                                      SongsList(SongsFile,Cache,False);{Songs List}
                                      ChangeSong(SongsFile,Cache,Song,TSongs);{Change active song}
                                      { i represents number of songs in file}
                                      {ChangeSong updates cache & song structures}
                                      CheckDisc;
                                   End
                                else
                                   Run:=False;


                 'M': If Valid[10] then            {Import Discs}
                         Begin
                            ImportDiscs(DiscsFile);
                            VDiscs;
                            CheckCD;
                         End
                      else
                         Run:=False;

                 'I': If Valid[11] then            {Import Songs}
                         Begin
                            ImportSongs;
                            VSongs(True);
                            CheckCD;
                         End
                      Else
                         Run:=False;

                   'R':
                      If Valid[12] then
                         Begin
                            ClrScr;

                            Write('Eject Door (Y/N) ?');
                            CursorOn;
                            Repeat
                                Key:=UpCase(ReadKey);
                                If Key=#0 then ReadKey;
                            Until Key in YesNo;
                            CursorOff;
                            Writeln;

                            If Key=Yes then
                               Begin
                                  If Locked then
                                     Begin
                                        Write('Unlocking Tray... ');
                                        UnLockDoor;
                                        Check(Not Locked,False);
                                     End;
                                                                                     Writeln('Ejecting Tray...');
                                  EjectDoor;
                                  While (DrvBusy or Not DoorOpen) and (Not Keypressed) do;
                                  Writeln('Change Discs and press any key to continue.');
                                  Pause;
                               End;

                            CD:=True;
                            InitCD;

                            If CD then
                               Begin
                                  Menu[NOE,1]:=' <R>   = Rescan Disc in drive, or add to database';
                                  Menu[NOE,2]:='Rescan CD drive and add if not in database.';
                                  ClrScr;
                                  CheckCD;        {Rescan/Add Disc in drive}
                               End;
                         End
                           Else
                              Run:=False;
                  #255:                                  {Credits}
                      Begin
                         ClrScr;
                         CursorOn;
                         Writeln(Copyright);
                         Writeln;
                         Writeln('Distribution license : GNU''s GPL ');
                         Writeln('Compiler : Borland Turbo Pascal 7.0');
                         Writeln('Check the attached files for more details.');
                         Writeln;
                         Writeln('Many hours of sweat were invested in this pseudo-small program.');
                         Writeln;
                         Writeln('*Please publicize your code so other people can learn from it*');
                         Writeln;
                         GotoXY(1,25);
                         Write('Press any key to return to menu...');

                         Pause;
                     End;
              Else
                 Run:=False;
              End;
         Until Run;
    Until Key=#27;
    Finish(True);  {Call our semi-exit procedure}
end.
(* End of file *)

--------------------------------------------
Parameters :
/D - Disables CD-SUPPORT
--------------------------------------------------------------
